@@ -64,6 +64,8 @@ struct wipe_desc {
 struct wipe_control {
 	char		*devname;
 	const char	*type_pattern;		/* -t <pattern> */
+	const char	*lockmode;
+	const char	*backup;		/* location of backups */
 
 	struct libscols_table *outtab;
 	struct wipe_desc *offsets;		/* -o <offset> -o <offset> ... */
@@ -76,7 +78,6 @@ struct wipe_control {
 	unsigned int	noact : 1,
 			all : 1,
 			quiet : 1,
-			backup : 1,
 			force : 1,
 			json : 1,
 			no_headings : 1,
@@ -187,28 +188,6 @@ static void init_output(struct wipe_control *ctl)
 
 static void finalize_output(struct wipe_control *ctl)
 {
-	if (ctl->parsable && !ctl->no_headings
-	    && !scols_table_is_empty(ctl->outtab)) {
-		struct libscols_iter *itr = scols_new_iter(SCOLS_ITER_FORWARD);
-		struct libscols_column *cl;
-		int i = 0;
-
-		if (!itr)
-			err_oom();
-
-		fputs("# ", stdout);
-		while (scols_table_next_column(ctl->outtab, itr, &cl) == 0) {
-			struct libscols_cell *hdr = scols_column_get_header(cl);
-			const char *name = scols_cell_get_data(hdr);
-
-			if (i)
-				fputc(',', stdout);
-			fputs(name, stdout);
-			i++;
-		}
-		fputc('\n', stdout);
-		scols_free_iter(itr);
-	}
 	scols_print_table(ctl->outtab);
 	scols_unref_table(ctl->outtab);
 }
@@ -307,7 +286,7 @@ static struct wipe_desc *get_desc_for_probe(struct wipe_control *ctl,
 					    loff_t *offset,
 					    size_t *len)
 {
-	const char *off, *type, *mag, *p, *usage = NULL;
+	const char *off, *type, *mag, *p, *use = NULL;
 	struct wipe_desc *wp;
 	int rc, ispt = 0;
 
@@ -328,12 +307,15 @@ static struct wipe_desc *get_desc_for_probe(struct wipe_control *ctl,
 			rc = blkid_probe_lookup_value(pr, "PTMAGIC", &mag, len);
 		if (rc)
 			return NULL;
-		usage = N_("partition-table");
+		use = N_("partition-table");
 		ispt = 1;
 	} else
 		return NULL;
 
+	errno = 0;
 	*offset = strtoll(off, NULL, 10);
+	if (errno)
+		return NULL;
 
 	/* Filter out by -t <type> */
 	if (ctl->type_pattern && !match_fstype(type, ctl->type_pattern))
@@ -357,8 +339,8 @@ static struct wipe_desc *get_desc_for_probe(struct wipe_control *ctl,
 	if (!wp)
 		return NULL;
 
-	if (usage || blkid_probe_lookup_value(pr, "USAGE", &usage, NULL) == 0)
-		wp->usage = xstrdup(usage);
+	if (use || blkid_probe_lookup_value(pr, "USAGE", &use, NULL) == 0)
+		wp->usage = xstrdup(use);
 
 	wp->type = xstrdup(type);
 	wp->on_disk = 1;
@@ -386,7 +368,7 @@ new_probe(const char *devname, int mode)
 		return NULL;
 
 	if (mode) {
-		int fd = open(devname, mode);
+		int fd = open(devname, mode | O_NONBLOCK);
 		if (fd < 0)
 			goto error;
 
@@ -520,7 +502,7 @@ static void rereadpt(int fd, const char *devname)
 		 * without delay is uncuccesful. The reason is probably kernel
 		 * and/or udevd.  Let's wait a moment and try more attempts.
 		 */
-		xusleep(25000);
+		xusleep(250000);
 
 		errno = 0;
 		ioctl(fd, BLKRRPART);
@@ -546,13 +528,16 @@ static int do_wipe(struct wipe_control *ctl)
 	if (!pr)
 		return -errno;
 
+	if (blkdev_lock(blkid_probe_get_fd(pr),
+			ctl->devname, ctl->lockmode) != 0) {
+		blkid_free_probe(pr);
+		return -1;
+	}
+
 	if (ctl->backup) {
-		const char *home = getenv ("HOME");
 		char *tmp = xstrdup(ctl->devname);
 
-		if (!home)
-			errx(EXIT_FAILURE, _("failed to create a signature backup, $HOME undefined"));
-		xasprintf (&backup, "%s/wipefs-%s-", home, basename(tmp));
+		xasprintf(&backup, "%s/wipefs-%s-", ctl->backup, basename(tmp));
 		free(tmp);
 	}
 
@@ -605,7 +590,9 @@ static int do_wipe(struct wipe_control *ctl)
 	if (need_force)
 		warnx(_("Use the --force option to force erase."));
 
-	fsync(blkid_probe_get_fd(pr));
+	if (fsync(blkid_probe_get_fd(pr)) != 0)
+		err(EXIT_FAILURE, _("%s: cannot flush modified buffers"),
+				ctl->devname);
 
 #ifdef BLKRRPART
 	if (reread && (mode & O_EXCL)) {
@@ -625,7 +612,9 @@ static int do_wipe(struct wipe_control *ctl)
 	}
 #endif
 
-	close(blkid_probe_get_fd(pr));
+	if (close(blkid_probe_get_fd(pr)) != 0)
+		err(EXIT_FAILURE, _("%s: close device failed"), ctl->devname);
+
 	blkid_free_probe(pr);
 	free(backup);
 	return 0;
@@ -638,31 +627,36 @@ usage(void)
 	size_t i;
 
 	fputs(USAGE_HEADER, stdout);
-	printf(_(" %s [options] <device>\n"), program_invocation_short_name);
+	fprintf(stdout, _(" %s [options] <device>\n"), program_invocation_short_name);
 
 	fputs(USAGE_SEPARATOR, stdout);
-	puts(_("Wipe signatures from a device."));
+	fputsln(_("Wipe signatures from a device."), stdout);
 
 	fputs(USAGE_OPTIONS, stdout);
-	puts(_(" -a, --all           wipe all magic strings (BE CAREFUL!)"));
-	puts(_(" -b, --backup        create a signature backup in $HOME"));
-	puts(_(" -f, --force         force erasure"));
-	puts(_(" -i, --noheadings    don't print headings"));
-	puts(_(" -J, --json          use JSON output format"));
-	puts(_(" -n, --no-act        do everything except the actual write() call"));
-	puts(_(" -o, --offset <num>  offset to erase, in bytes"));
-	puts(_(" -O, --output <list> COLUMNS to display (see below)"));
-	puts(_(" -p, --parsable      print out in parsable instead of printable format"));
-	puts(_(" -q, --quiet         suppress output messages"));
-	puts(_(" -t, --types <list>  limit the set of filesystem, RAIDs or partition tables"));
+	fputsln(_(" -a, --all            wipe all magic strings (BE CAREFUL!)"), stdout);
+	fputsln(_(" -b, --backup[=<dir>] create a signature backup in <dir> or $HOME"), stdout);
+	fputsln(_(" -f, --force          force erasure"), stdout);
+	fputsln(_(" -i, --noheadings     don't print headings"), stdout);
+	fputsln(_(" -J, --json           use JSON output format"), stdout);
+	fputsln(_(" -n, --no-act         do everything except the actual write() call"), stdout);
+	fputsln(_(" -o, --offset <num>   offset to erase, in bytes"), stdout);
+	fputsln(_(" -O, --output <list>  COLUMNS to display (see below)"), stdout);
+	fputsln(_(" -p, --parsable       print out in parsable instead of printable format"), stdout);
+	fputsln(_(" -q, --quiet          suppress output messages"), stdout);
+	fputsln(_(" -t, --types <list>   limit the set of filesystem, RAIDs or partition tables"), stdout);
+	fprintf(stdout,
+	     _("     --lock[=<mode>] use exclusive device lock (%s, %s or %s)\n"), "yes", "no", "nonblock");
 
-	printf(USAGE_HELP_OPTIONS(21));
+	fprintf(stdout, USAGE_HELP_OPTIONS(22));
+
+	fputs(USAGE_ARGUMENTS, stdout);
+	fprintf(stdout, USAGE_ARG_SIZE(_("<num>")));
 
 	fputs(USAGE_COLUMNS, stdout);
 	for (i = 0; i < ARRAY_SIZE(infos); i++)
 		fprintf(stdout, " %8s  %s\n", infos[i].name, _(infos[i].help));
 
-	printf(USAGE_MAN_TAIL("wipefs(8)"));
+	fprintf(stdout, USAGE_MAN_TAIL("wipefs(8)"));
 	exit(EXIT_SUCCESS);
 }
 
@@ -672,13 +666,17 @@ main(int argc, char **argv)
 {
 	struct wipe_control ctl = { .devname = NULL };
 	int c;
+	size_t i;
 	char *outarg = NULL;
-
+	enum {
+		OPT_LOCK = CHAR_MAX + 1,
+	};
 	static const struct option longopts[] = {
 	    { "all",       no_argument,       NULL, 'a' },
-	    { "backup",    no_argument,       NULL, 'b' },
+	    { "backup",    optional_argument, NULL, 'b' },
 	    { "force",     no_argument,       NULL, 'f' },
 	    { "help",      no_argument,       NULL, 'h' },
+	    { "lock",      optional_argument, NULL, OPT_LOCK },
 	    { "no-act",    no_argument,       NULL, 'n' },
 	    { "offset",    required_argument, NULL, 'o' },
 	    { "parsable",  no_argument,       NULL, 'p' },
@@ -700,9 +698,9 @@ main(int argc, char **argv)
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
-	atexit(close_stdout);
+	close_stdout_atexit();
 
-	while ((c = getopt_long(argc, argv, "abfhiJnO:o:pqt:V", longopts, NULL)) != -1) {
+	while ((c = getopt_long(argc, argv, "ab::fhiJnO:o:pqt:V", longopts, NULL)) != -1) {
 
 		err_exclusive_options(c, longopts, excl, excl_st);
 
@@ -711,14 +709,18 @@ main(int argc, char **argv)
 			ctl.all = 1;
 			break;
 		case 'b':
-			ctl.backup = 1;
+			if (optarg) {
+				ctl.backup = optarg;
+			} else {
+				ctl.backup = getenv("HOME");
+				if (!ctl.backup)
+					errx(EXIT_FAILURE,
+					     _("failed to create a signature backup, $HOME undefined"));
+			}
 			break;
 		case 'f':
 			ctl.force = 1;
 			break;
-		case 'h':
-			usage();
-			break;
 		case 'J':
 			ctl.json = 1;
 			break;
@@ -745,9 +747,18 @@ main(int argc, char **argv)
 		case 't':
 			ctl.type_pattern = optarg;
 			break;
+		case OPT_LOCK:
+			ctl.lockmode = "1";
+			if (optarg) {
+				if (*optarg == '=')
+					optarg++;
+				ctl.lockmode = optarg;
+			}
+			break;
+		case 'h':
+			usage();
 		case 'V':
-			printf(UTIL_LINUX_VERSION);
-			return EXIT_SUCCESS;
+			print_version(EXIT_SUCCESS);
 		default:
 			errtryhelp(EXIT_FAILURE);
 		}
@@ -814,7 +825,7 @@ main(int argc, char **argv)
 		/* Re-read partition tables on whole-disk devices. This is
 		 * postponed until all is done to avoid conflicts.
 		 */
-		for (size_t i = 0; i < ctl.nrereads; i++) {
+		for (i = 0; i < ctl.nrereads; i++) {
 			char *devname = ctl.reread[i];
 			int fd = open(devname, O_RDONLY);
 
